import numpy
from typing import Union, TypeVar
from functools import wraps
from .generic_utils import container_catamorphism, type_match


USE_QUICK_ATTRIBUTES = True


class GraphicsNDArray(numpy.ndarray):
    """
    A subclass of ``numpy.ndarray``.

    It adds GLSL-like access to components,
    e.g. ``a.xyz / a.w`` is a shorthand for ``a[..., [0, 1, 2]] / a[..., [3]]``.

    Note that single-component results retains a singleton ``1`` dimension.
    """

    def __new__(cls, input_array):
        return numpy.asarray(input_array).view(cls)

    def __array_function__(self, func, types, args, kwargs):
        result = super().__array_function__(func, types, args, kwargs)
        if isinstance(result, numpy.ndarray):
            # print("__array_function__ cast")
            return GraphicsNDArray(result)
        return result

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):  # this method is called whenever you use a ufunc
        f = {
            "reduce": ufunc.reduce,
            "accumulate": ufunc.accumulate,
            "reduceat": ufunc.reduceat,
            "outer": ufunc.outer,
            "at": ufunc.at,
            "__call__": ufunc,
        }
        # print("__array_ufunc__ cast")
        output = GraphicsNDArray(f[method](*(numpy.asarray(i) for i in inputs), **kwargs))
        return output

    def qattr(x: tuple):
        return property(lambda self: self[x])

    if USE_QUICK_ATTRIBUTES:
        x: "GraphicsNDArray" = qattr((..., [0]))
        y: "GraphicsNDArray" = qattr((..., [1]))
        z: "GraphicsNDArray" = qattr((..., [2]))
        w: "GraphicsNDArray" = qattr((..., [3]))
        xx: "GraphicsNDArray" = qattr((..., [0, 0]))
        xy: "GraphicsNDArray" = qattr((..., [0, 1]))
        xz: "GraphicsNDArray" = qattr((..., [0, 2]))
        xw: "GraphicsNDArray" = qattr((..., [0, 3]))
        yx: "GraphicsNDArray" = qattr((..., [1, 0]))
        yy: "GraphicsNDArray" = qattr((..., [1, 1]))
        yz: "GraphicsNDArray" = qattr((..., [1, 2]))
        yw: "GraphicsNDArray" = qattr((..., [1, 3]))
        zx: "GraphicsNDArray" = qattr((..., [2, 0]))
        zy: "GraphicsNDArray" = qattr((..., [2, 1]))
        zz: "GraphicsNDArray" = qattr((..., [2, 2]))
        zw: "GraphicsNDArray" = qattr((..., [2, 3]))
        wx: "GraphicsNDArray" = qattr((..., [3, 0]))
        wy: "GraphicsNDArray" = qattr((..., [3, 1]))
        wz: "GraphicsNDArray" = qattr((..., [3, 2]))
        ww: "GraphicsNDArray" = qattr((..., [3, 3]))
        xxx: "GraphicsNDArray" = qattr((..., [0, 0, 0]))
        xxy: "GraphicsNDArray" = qattr((..., [0, 0, 1]))
        xxz: "GraphicsNDArray" = qattr((..., [0, 0, 2]))
        xxw: "GraphicsNDArray" = qattr((..., [0, 0, 3]))
        xyx: "GraphicsNDArray" = qattr((..., [0, 1, 0]))
        xyy: "GraphicsNDArray" = qattr((..., [0, 1, 1]))
        xyz: "GraphicsNDArray" = qattr((..., [0, 1, 2]))
        xyw: "GraphicsNDArray" = qattr((..., [0, 1, 3]))
        xzx: "GraphicsNDArray" = qattr((..., [0, 2, 0]))
        xzy: "GraphicsNDArray" = qattr((..., [0, 2, 1]))
        xzz: "GraphicsNDArray" = qattr((..., [0, 2, 2]))
        xzw: "GraphicsNDArray" = qattr((..., [0, 2, 3]))
        xwx: "GraphicsNDArray" = qattr((..., [0, 3, 0]))
        xwy: "GraphicsNDArray" = qattr((..., [0, 3, 1]))
        xwz: "GraphicsNDArray" = qattr((..., [0, 3, 2]))
        xww: "GraphicsNDArray" = qattr((..., [0, 3, 3]))
        yxx: "GraphicsNDArray" = qattr((..., [1, 0, 0]))
        yxy: "GraphicsNDArray" = qattr((..., [1, 0, 1]))
        yxz: "GraphicsNDArray" = qattr((..., [1, 0, 2]))
        yxw: "GraphicsNDArray" = qattr((..., [1, 0, 3]))
        yyx: "GraphicsNDArray" = qattr((..., [1, 1, 0]))
        yyy: "GraphicsNDArray" = qattr((..., [1, 1, 1]))
        yyz: "GraphicsNDArray" = qattr((..., [1, 1, 2]))
        yyw: "GraphicsNDArray" = qattr((..., [1, 1, 3]))
        yzx: "GraphicsNDArray" = qattr((..., [1, 2, 0]))
        yzy: "GraphicsNDArray" = qattr((..., [1, 2, 1]))
        yzz: "GraphicsNDArray" = qattr((..., [1, 2, 2]))
        yzw: "GraphicsNDArray" = qattr((..., [1, 2, 3]))
        ywx: "GraphicsNDArray" = qattr((..., [1, 3, 0]))
        ywy: "GraphicsNDArray" = qattr((..., [1, 3, 1]))
        ywz: "GraphicsNDArray" = qattr((..., [1, 3, 2]))
        yww: "GraphicsNDArray" = qattr((..., [1, 3, 3]))
        zxx: "GraphicsNDArray" = qattr((..., [2, 0, 0]))
        zxy: "GraphicsNDArray" = qattr((..., [2, 0, 1]))
        zxz: "GraphicsNDArray" = qattr((..., [2, 0, 2]))
        zxw: "GraphicsNDArray" = qattr((..., [2, 0, 3]))
        zyx: "GraphicsNDArray" = qattr((..., [2, 1, 0]))
        zyy: "GraphicsNDArray" = qattr((..., [2, 1, 1]))
        zyz: "GraphicsNDArray" = qattr((..., [2, 1, 2]))
        zyw: "GraphicsNDArray" = qattr((..., [2, 1, 3]))
        zzx: "GraphicsNDArray" = qattr((..., [2, 2, 0]))
        zzy: "GraphicsNDArray" = qattr((..., [2, 2, 1]))
        zzz: "GraphicsNDArray" = qattr((..., [2, 2, 2]))
        zzw: "GraphicsNDArray" = qattr((..., [2, 2, 3]))
        zwx: "GraphicsNDArray" = qattr((..., [2, 3, 0]))
        zwy: "GraphicsNDArray" = qattr((..., [2, 3, 1]))
        zwz: "GraphicsNDArray" = qattr((..., [2, 3, 2]))
        zww: "GraphicsNDArray" = qattr((..., [2, 3, 3]))
        wxx: "GraphicsNDArray" = qattr((..., [3, 0, 0]))
        wxy: "GraphicsNDArray" = qattr((..., [3, 0, 1]))
        wxz: "GraphicsNDArray" = qattr((..., [3, 0, 2]))
        wxw: "GraphicsNDArray" = qattr((..., [3, 0, 3]))
        wyx: "GraphicsNDArray" = qattr((..., [3, 1, 0]))
        wyy: "GraphicsNDArray" = qattr((..., [3, 1, 1]))
        wyz: "GraphicsNDArray" = qattr((..., [3, 1, 2]))
        wyw: "GraphicsNDArray" = qattr((..., [3, 1, 3]))
        wzx: "GraphicsNDArray" = qattr((..., [3, 2, 0]))
        wzy: "GraphicsNDArray" = qattr((..., [3, 2, 1]))
        wzz: "GraphicsNDArray" = qattr((..., [3, 2, 2]))
        wzw: "GraphicsNDArray" = qattr((..., [3, 2, 3]))
        wwx: "GraphicsNDArray" = qattr((..., [3, 3, 0]))
        wwy: "GraphicsNDArray" = qattr((..., [3, 3, 1]))
        wwz: "GraphicsNDArray" = qattr((..., [3, 3, 2]))
        www: "GraphicsNDArray" = qattr((..., [3, 3, 3]))
        xxxx: "GraphicsNDArray" = qattr((..., [0, 0, 0, 0]))
        xxxy: "GraphicsNDArray" = qattr((..., [0, 0, 0, 1]))
        xxxz: "GraphicsNDArray" = qattr((..., [0, 0, 0, 2]))
        xxxw: "GraphicsNDArray" = qattr((..., [0, 0, 0, 3]))
        xxyx: "GraphicsNDArray" = qattr((..., [0, 0, 1, 0]))
        xxyy: "GraphicsNDArray" = qattr((..., [0, 0, 1, 1]))
        xxyz: "GraphicsNDArray" = qattr((..., [0, 0, 1, 2]))
        xxyw: "GraphicsNDArray" = qattr((..., [0, 0, 1, 3]))
        xxzx: "GraphicsNDArray" = qattr((..., [0, 0, 2, 0]))
        xxzy: "GraphicsNDArray" = qattr((..., [0, 0, 2, 1]))
        xxzz: "GraphicsNDArray" = qattr((..., [0, 0, 2, 2]))
        xxzw: "GraphicsNDArray" = qattr((..., [0, 0, 2, 3]))
        xxwx: "GraphicsNDArray" = qattr((..., [0, 0, 3, 0]))
        xxwy: "GraphicsNDArray" = qattr((..., [0, 0, 3, 1]))
        xxwz: "GraphicsNDArray" = qattr((..., [0, 0, 3, 2]))
        xxww: "GraphicsNDArray" = qattr((..., [0, 0, 3, 3]))
        xyxx: "GraphicsNDArray" = qattr((..., [0, 1, 0, 0]))
        xyxy: "GraphicsNDArray" = qattr((..., [0, 1, 0, 1]))
        xyxz: "GraphicsNDArray" = qattr((..., [0, 1, 0, 2]))
        xyxw: "GraphicsNDArray" = qattr((..., [0, 1, 0, 3]))
        xyyx: "GraphicsNDArray" = qattr((..., [0, 1, 1, 0]))
        xyyy: "GraphicsNDArray" = qattr((..., [0, 1, 1, 1]))
        xyyz: "GraphicsNDArray" = qattr((..., [0, 1, 1, 2]))
        xyyw: "GraphicsNDArray" = qattr((..., [0, 1, 1, 3]))
        xyzx: "GraphicsNDArray" = qattr((..., [0, 1, 2, 0]))
        xyzy: "GraphicsNDArray" = qattr((..., [0, 1, 2, 1]))
        xyzz: "GraphicsNDArray" = qattr((..., [0, 1, 2, 2]))
        xyzw: "GraphicsNDArray" = qattr((..., [0, 1, 2, 3]))
        xywx: "GraphicsNDArray" = qattr((..., [0, 1, 3, 0]))
        xywy: "GraphicsNDArray" = qattr((..., [0, 1, 3, 1]))
        xywz: "GraphicsNDArray" = qattr((..., [0, 1, 3, 2]))
        xyww: "GraphicsNDArray" = qattr((..., [0, 1, 3, 3]))
        xzxx: "GraphicsNDArray" = qattr((..., [0, 2, 0, 0]))
        xzxy: "GraphicsNDArray" = qattr((..., [0, 2, 0, 1]))
        xzxz: "GraphicsNDArray" = qattr((..., [0, 2, 0, 2]))
        xzxw: "GraphicsNDArray" = qattr((..., [0, 2, 0, 3]))
        xzyx: "GraphicsNDArray" = qattr((..., [0, 2, 1, 0]))
        xzyy: "GraphicsNDArray" = qattr((..., [0, 2, 1, 1]))
        xzyz: "GraphicsNDArray" = qattr((..., [0, 2, 1, 2]))
        xzyw: "GraphicsNDArray" = qattr((..., [0, 2, 1, 3]))
        xzzx: "GraphicsNDArray" = qattr((..., [0, 2, 2, 0]))
        xzzy: "GraphicsNDArray" = qattr((..., [0, 2, 2, 1]))
        xzzz: "GraphicsNDArray" = qattr((..., [0, 2, 2, 2]))
        xzzw: "GraphicsNDArray" = qattr((..., [0, 2, 2, 3]))
        xzwx: "GraphicsNDArray" = qattr((..., [0, 2, 3, 0]))
        xzwy: "GraphicsNDArray" = qattr((..., [0, 2, 3, 1]))
        xzwz: "GraphicsNDArray" = qattr((..., [0, 2, 3, 2]))
        xzww: "GraphicsNDArray" = qattr((..., [0, 2, 3, 3]))
        xwxx: "GraphicsNDArray" = qattr((..., [0, 3, 0, 0]))
        xwxy: "GraphicsNDArray" = qattr((..., [0, 3, 0, 1]))
        xwxz: "GraphicsNDArray" = qattr((..., [0, 3, 0, 2]))
        xwxw: "GraphicsNDArray" = qattr((..., [0, 3, 0, 3]))
        xwyx: "GraphicsNDArray" = qattr((..., [0, 3, 1, 0]))
        xwyy: "GraphicsNDArray" = qattr((..., [0, 3, 1, 1]))
        xwyz: "GraphicsNDArray" = qattr((..., [0, 3, 1, 2]))
        xwyw: "GraphicsNDArray" = qattr((..., [0, 3, 1, 3]))
        xwzx: "GraphicsNDArray" = qattr((..., [0, 3, 2, 0]))
        xwzy: "GraphicsNDArray" = qattr((..., [0, 3, 2, 1]))
        xwzz: "GraphicsNDArray" = qattr((..., [0, 3, 2, 2]))
        xwzw: "GraphicsNDArray" = qattr((..., [0, 3, 2, 3]))
        xwwx: "GraphicsNDArray" = qattr((..., [0, 3, 3, 0]))
        xwwy: "GraphicsNDArray" = qattr((..., [0, 3, 3, 1]))
        xwwz: "GraphicsNDArray" = qattr((..., [0, 3, 3, 2]))
        xwww: "GraphicsNDArray" = qattr((..., [0, 3, 3, 3]))
        yxxx: "GraphicsNDArray" = qattr((..., [1, 0, 0, 0]))
        yxxy: "GraphicsNDArray" = qattr((..., [1, 0, 0, 1]))
        yxxz: "GraphicsNDArray" = qattr((..., [1, 0, 0, 2]))
        yxxw: "GraphicsNDArray" = qattr((..., [1, 0, 0, 3]))
        yxyx: "GraphicsNDArray" = qattr((..., [1, 0, 1, 0]))
        yxyy: "GraphicsNDArray" = qattr((..., [1, 0, 1, 1]))
        yxyz: "GraphicsNDArray" = qattr((..., [1, 0, 1, 2]))
        yxyw: "GraphicsNDArray" = qattr((..., [1, 0, 1, 3]))
        yxzx: "GraphicsNDArray" = qattr((..., [1, 0, 2, 0]))
        yxzy: "GraphicsNDArray" = qattr((..., [1, 0, 2, 1]))
        yxzz: "GraphicsNDArray" = qattr((..., [1, 0, 2, 2]))
        yxzw: "GraphicsNDArray" = qattr((..., [1, 0, 2, 3]))
        yxwx: "GraphicsNDArray" = qattr((..., [1, 0, 3, 0]))
        yxwy: "GraphicsNDArray" = qattr((..., [1, 0, 3, 1]))
        yxwz: "GraphicsNDArray" = qattr((..., [1, 0, 3, 2]))
        yxww: "GraphicsNDArray" = qattr((..., [1, 0, 3, 3]))
        yyxx: "GraphicsNDArray" = qattr((..., [1, 1, 0, 0]))
        yyxy: "GraphicsNDArray" = qattr((..., [1, 1, 0, 1]))
        yyxz: "GraphicsNDArray" = qattr((..., [1, 1, 0, 2]))
        yyxw: "GraphicsNDArray" = qattr((..., [1, 1, 0, 3]))
        yyyx: "GraphicsNDArray" = qattr((..., [1, 1, 1, 0]))
        yyyy: "GraphicsNDArray" = qattr((..., [1, 1, 1, 1]))
        yyyz: "GraphicsNDArray" = qattr((..., [1, 1, 1, 2]))
        yyyw: "GraphicsNDArray" = qattr((..., [1, 1, 1, 3]))
        yyzx: "GraphicsNDArray" = qattr((..., [1, 1, 2, 0]))
        yyzy: "GraphicsNDArray" = qattr((..., [1, 1, 2, 1]))
        yyzz: "GraphicsNDArray" = qattr((..., [1, 1, 2, 2]))
        yyzw: "GraphicsNDArray" = qattr((..., [1, 1, 2, 3]))
        yywx: "GraphicsNDArray" = qattr((..., [1, 1, 3, 0]))
        yywy: "GraphicsNDArray" = qattr((..., [1, 1, 3, 1]))
        yywz: "GraphicsNDArray" = qattr((..., [1, 1, 3, 2]))
        yyww: "GraphicsNDArray" = qattr((..., [1, 1, 3, 3]))
        yzxx: "GraphicsNDArray" = qattr((..., [1, 2, 0, 0]))
        yzxy: "GraphicsNDArray" = qattr((..., [1, 2, 0, 1]))
        yzxz: "GraphicsNDArray" = qattr((..., [1, 2, 0, 2]))
        yzxw: "GraphicsNDArray" = qattr((..., [1, 2, 0, 3]))
        yzyx: "GraphicsNDArray" = qattr((..., [1, 2, 1, 0]))
        yzyy: "GraphicsNDArray" = qattr((..., [1, 2, 1, 1]))
        yzyz: "GraphicsNDArray" = qattr((..., [1, 2, 1, 2]))
        yzyw: "GraphicsNDArray" = qattr((..., [1, 2, 1, 3]))
        yzzx: "GraphicsNDArray" = qattr((..., [1, 2, 2, 0]))
        yzzy: "GraphicsNDArray" = qattr((..., [1, 2, 2, 1]))
        yzzz: "GraphicsNDArray" = qattr((..., [1, 2, 2, 2]))
        yzzw: "GraphicsNDArray" = qattr((..., [1, 2, 2, 3]))
        yzwx: "GraphicsNDArray" = qattr((..., [1, 2, 3, 0]))
        yzwy: "GraphicsNDArray" = qattr((..., [1, 2, 3, 1]))
        yzwz: "GraphicsNDArray" = qattr((..., [1, 2, 3, 2]))
        yzww: "GraphicsNDArray" = qattr((..., [1, 2, 3, 3]))
        ywxx: "GraphicsNDArray" = qattr((..., [1, 3, 0, 0]))
        ywxy: "GraphicsNDArray" = qattr((..., [1, 3, 0, 1]))
        ywxz: "GraphicsNDArray" = qattr((..., [1, 3, 0, 2]))
        ywxw: "GraphicsNDArray" = qattr((..., [1, 3, 0, 3]))
        ywyx: "GraphicsNDArray" = qattr((..., [1, 3, 1, 0]))
        ywyy: "GraphicsNDArray" = qattr((..., [1, 3, 1, 1]))
        ywyz: "GraphicsNDArray" = qattr((..., [1, 3, 1, 2]))
        ywyw: "GraphicsNDArray" = qattr((..., [1, 3, 1, 3]))
        ywzx: "GraphicsNDArray" = qattr((..., [1, 3, 2, 0]))
        ywzy: "GraphicsNDArray" = qattr((..., [1, 3, 2, 1]))
        ywzz: "GraphicsNDArray" = qattr((..., [1, 3, 2, 2]))
        ywzw: "GraphicsNDArray" = qattr((..., [1, 3, 2, 3]))
        ywwx: "GraphicsNDArray" = qattr((..., [1, 3, 3, 0]))
        ywwy: "GraphicsNDArray" = qattr((..., [1, 3, 3, 1]))
        ywwz: "GraphicsNDArray" = qattr((..., [1, 3, 3, 2]))
        ywww: "GraphicsNDArray" = qattr((..., [1, 3, 3, 3]))
        zxxx: "GraphicsNDArray" = qattr((..., [2, 0, 0, 0]))
        zxxy: "GraphicsNDArray" = qattr((..., [2, 0, 0, 1]))
        zxxz: "GraphicsNDArray" = qattr((..., [2, 0, 0, 2]))
        zxxw: "GraphicsNDArray" = qattr((..., [2, 0, 0, 3]))
        zxyx: "GraphicsNDArray" = qattr((..., [2, 0, 1, 0]))
        zxyy: "GraphicsNDArray" = qattr((..., [2, 0, 1, 1]))
        zxyz: "GraphicsNDArray" = qattr((..., [2, 0, 1, 2]))
        zxyw: "GraphicsNDArray" = qattr((..., [2, 0, 1, 3]))
        zxzx: "GraphicsNDArray" = qattr((..., [2, 0, 2, 0]))
        zxzy: "GraphicsNDArray" = qattr((..., [2, 0, 2, 1]))
        zxzz: "GraphicsNDArray" = qattr((..., [2, 0, 2, 2]))
        zxzw: "GraphicsNDArray" = qattr((..., [2, 0, 2, 3]))
        zxwx: "GraphicsNDArray" = qattr((..., [2, 0, 3, 0]))
        zxwy: "GraphicsNDArray" = qattr((..., [2, 0, 3, 1]))
        zxwz: "GraphicsNDArray" = qattr((..., [2, 0, 3, 2]))
        zxww: "GraphicsNDArray" = qattr((..., [2, 0, 3, 3]))
        zyxx: "GraphicsNDArray" = qattr((..., [2, 1, 0, 0]))
        zyxy: "GraphicsNDArray" = qattr((..., [2, 1, 0, 1]))
        zyxz: "GraphicsNDArray" = qattr((..., [2, 1, 0, 2]))
        zyxw: "GraphicsNDArray" = qattr((..., [2, 1, 0, 3]))
        zyyx: "GraphicsNDArray" = qattr((..., [2, 1, 1, 0]))
        zyyy: "GraphicsNDArray" = qattr((..., [2, 1, 1, 1]))
        zyyz: "GraphicsNDArray" = qattr((..., [2, 1, 1, 2]))
        zyyw: "GraphicsNDArray" = qattr((..., [2, 1, 1, 3]))
        zyzx: "GraphicsNDArray" = qattr((..., [2, 1, 2, 0]))
        zyzy: "GraphicsNDArray" = qattr((..., [2, 1, 2, 1]))
        zyzz: "GraphicsNDArray" = qattr((..., [2, 1, 2, 2]))
        zyzw: "GraphicsNDArray" = qattr((..., [2, 1, 2, 3]))
        zywx: "GraphicsNDArray" = qattr((..., [2, 1, 3, 0]))
        zywy: "GraphicsNDArray" = qattr((..., [2, 1, 3, 1]))
        zywz: "GraphicsNDArray" = qattr((..., [2, 1, 3, 2]))
        zyww: "GraphicsNDArray" = qattr((..., [2, 1, 3, 3]))
        zzxx: "GraphicsNDArray" = qattr((..., [2, 2, 0, 0]))
        zzxy: "GraphicsNDArray" = qattr((..., [2, 2, 0, 1]))
        zzxz: "GraphicsNDArray" = qattr((..., [2, 2, 0, 2]))
        zzxw: "GraphicsNDArray" = qattr((..., [2, 2, 0, 3]))
        zzyx: "GraphicsNDArray" = qattr((..., [2, 2, 1, 0]))
        zzyy: "GraphicsNDArray" = qattr((..., [2, 2, 1, 1]))
        zzyz: "GraphicsNDArray" = qattr((..., [2, 2, 1, 2]))
        zzyw: "GraphicsNDArray" = qattr((..., [2, 2, 1, 3]))
        zzzx: "GraphicsNDArray" = qattr((..., [2, 2, 2, 0]))
        zzzy: "GraphicsNDArray" = qattr((..., [2, 2, 2, 1]))
        zzzz: "GraphicsNDArray" = qattr((..., [2, 2, 2, 2]))
        zzzw: "GraphicsNDArray" = qattr((..., [2, 2, 2, 3]))
        zzwx: "GraphicsNDArray" = qattr((..., [2, 2, 3, 0]))
        zzwy: "GraphicsNDArray" = qattr((..., [2, 2, 3, 1]))
        zzwz: "GraphicsNDArray" = qattr((..., [2, 2, 3, 2]))
        zzww: "GraphicsNDArray" = qattr((..., [2, 2, 3, 3]))
        zwxx: "GraphicsNDArray" = qattr((..., [2, 3, 0, 0]))
        zwxy: "GraphicsNDArray" = qattr((..., [2, 3, 0, 1]))
        zwxz: "GraphicsNDArray" = qattr((..., [2, 3, 0, 2]))
        zwxw: "GraphicsNDArray" = qattr((..., [2, 3, 0, 3]))
        zwyx: "GraphicsNDArray" = qattr((..., [2, 3, 1, 0]))
        zwyy: "GraphicsNDArray" = qattr((..., [2, 3, 1, 1]))
        zwyz: "GraphicsNDArray" = qattr((..., [2, 3, 1, 2]))
        zwyw: "GraphicsNDArray" = qattr((..., [2, 3, 1, 3]))
        zwzx: "GraphicsNDArray" = qattr((..., [2, 3, 2, 0]))
        zwzy: "GraphicsNDArray" = qattr((..., [2, 3, 2, 1]))
        zwzz: "GraphicsNDArray" = qattr((..., [2, 3, 2, 2]))
        zwzw: "GraphicsNDArray" = qattr((..., [2, 3, 2, 3]))
        zwwx: "GraphicsNDArray" = qattr((..., [2, 3, 3, 0]))
        zwwy: "GraphicsNDArray" = qattr((..., [2, 3, 3, 1]))
        zwwz: "GraphicsNDArray" = qattr((..., [2, 3, 3, 2]))
        zwww: "GraphicsNDArray" = qattr((..., [2, 3, 3, 3]))
        wxxx: "GraphicsNDArray" = qattr((..., [3, 0, 0, 0]))
        wxxy: "GraphicsNDArray" = qattr((..., [3, 0, 0, 1]))
        wxxz: "GraphicsNDArray" = qattr((..., [3, 0, 0, 2]))
        wxxw: "GraphicsNDArray" = qattr((..., [3, 0, 0, 3]))
        wxyx: "GraphicsNDArray" = qattr((..., [3, 0, 1, 0]))
        wxyy: "GraphicsNDArray" = qattr((..., [3, 0, 1, 1]))
        wxyz: "GraphicsNDArray" = qattr((..., [3, 0, 1, 2]))
        wxyw: "GraphicsNDArray" = qattr((..., [3, 0, 1, 3]))
        wxzx: "GraphicsNDArray" = qattr((..., [3, 0, 2, 0]))
        wxzy: "GraphicsNDArray" = qattr((..., [3, 0, 2, 1]))
        wxzz: "GraphicsNDArray" = qattr((..., [3, 0, 2, 2]))
        wxzw: "GraphicsNDArray" = qattr((..., [3, 0, 2, 3]))
        wxwx: "GraphicsNDArray" = qattr((..., [3, 0, 3, 0]))
        wxwy: "GraphicsNDArray" = qattr((..., [3, 0, 3, 1]))
        wxwz: "GraphicsNDArray" = qattr((..., [3, 0, 3, 2]))
        wxww: "GraphicsNDArray" = qattr((..., [3, 0, 3, 3]))
        wyxx: "GraphicsNDArray" = qattr((..., [3, 1, 0, 0]))
        wyxy: "GraphicsNDArray" = qattr((..., [3, 1, 0, 1]))
        wyxz: "GraphicsNDArray" = qattr((..., [3, 1, 0, 2]))
        wyxw: "GraphicsNDArray" = qattr((..., [3, 1, 0, 3]))
        wyyx: "GraphicsNDArray" = qattr((..., [3, 1, 1, 0]))
        wyyy: "GraphicsNDArray" = qattr((..., [3, 1, 1, 1]))
        wyyz: "GraphicsNDArray" = qattr((..., [3, 1, 1, 2]))
        wyyw: "GraphicsNDArray" = qattr((..., [3, 1, 1, 3]))
        wyzx: "GraphicsNDArray" = qattr((..., [3, 1, 2, 0]))
        wyzy: "GraphicsNDArray" = qattr((..., [3, 1, 2, 1]))
        wyzz: "GraphicsNDArray" = qattr((..., [3, 1, 2, 2]))
        wyzw: "GraphicsNDArray" = qattr((..., [3, 1, 2, 3]))
        wywx: "GraphicsNDArray" = qattr((..., [3, 1, 3, 0]))
        wywy: "GraphicsNDArray" = qattr((..., [3, 1, 3, 1]))
        wywz: "GraphicsNDArray" = qattr((..., [3, 1, 3, 2]))
        wyww: "GraphicsNDArray" = qattr((..., [3, 1, 3, 3]))
        wzxx: "GraphicsNDArray" = qattr((..., [3, 2, 0, 0]))
        wzxy: "GraphicsNDArray" = qattr((..., [3, 2, 0, 1]))
        wzxz: "GraphicsNDArray" = qattr((..., [3, 2, 0, 2]))
        wzxw: "GraphicsNDArray" = qattr((..., [3, 2, 0, 3]))
        wzyx: "GraphicsNDArray" = qattr((..., [3, 2, 1, 0]))
        wzyy: "GraphicsNDArray" = qattr((..., [3, 2, 1, 1]))
        wzyz: "GraphicsNDArray" = qattr((..., [3, 2, 1, 2]))
        wzyw: "GraphicsNDArray" = qattr((..., [3, 2, 1, 3]))
        wzzx: "GraphicsNDArray" = qattr((..., [3, 2, 2, 0]))
        wzzy: "GraphicsNDArray" = qattr((..., [3, 2, 2, 1]))
        wzzz: "GraphicsNDArray" = qattr((..., [3, 2, 2, 2]))
        wzzw: "GraphicsNDArray" = qattr((..., [3, 2, 2, 3]))
        wzwx: "GraphicsNDArray" = qattr((..., [3, 2, 3, 0]))
        wzwy: "GraphicsNDArray" = qattr((..., [3, 2, 3, 1]))
        wzwz: "GraphicsNDArray" = qattr((..., [3, 2, 3, 2]))
        wzww: "GraphicsNDArray" = qattr((..., [3, 2, 3, 3]))
        wwxx: "GraphicsNDArray" = qattr((..., [3, 3, 0, 0]))
        wwxy: "GraphicsNDArray" = qattr((..., [3, 3, 0, 1]))
        wwxz: "GraphicsNDArray" = qattr((..., [3, 3, 0, 2]))
        wwxw: "GraphicsNDArray" = qattr((..., [3, 3, 0, 3]))
        wwyx: "GraphicsNDArray" = qattr((..., [3, 3, 1, 0]))
        wwyy: "GraphicsNDArray" = qattr((..., [3, 3, 1, 1]))
        wwyz: "GraphicsNDArray" = qattr((..., [3, 3, 1, 2]))
        wwyw: "GraphicsNDArray" = qattr((..., [3, 3, 1, 3]))
        wwzx: "GraphicsNDArray" = qattr((..., [3, 3, 2, 0]))
        wwzy: "GraphicsNDArray" = qattr((..., [3, 3, 2, 1]))
        wwzz: "GraphicsNDArray" = qattr((..., [3, 3, 2, 2]))
        wwzw: "GraphicsNDArray" = qattr((..., [3, 3, 2, 3]))
        wwwx: "GraphicsNDArray" = qattr((..., [3, 3, 3, 0]))
        wwwy: "GraphicsNDArray" = qattr((..., [3, 3, 3, 1]))
        wwwz: "GraphicsNDArray" = qattr((..., [3, 3, 3, 2]))
        wwww: "GraphicsNDArray" = qattr((..., [3, 3, 3, 3]))
        r: "GraphicsNDArray" = qattr((..., [0]))
        g: "GraphicsNDArray" = qattr((..., [1]))
        b: "GraphicsNDArray" = qattr((..., [2]))
        a: "GraphicsNDArray" = qattr((..., [3]))
        rr: "GraphicsNDArray" = qattr((..., [0, 0]))
        rg: "GraphicsNDArray" = qattr((..., [0, 1]))
        rb: "GraphicsNDArray" = qattr((..., [0, 2]))
        ra: "GraphicsNDArray" = qattr((..., [0, 3]))
        gr: "GraphicsNDArray" = qattr((..., [1, 0]))
        gg: "GraphicsNDArray" = qattr((..., [1, 1]))
        gb: "GraphicsNDArray" = qattr((..., [1, 2]))
        ga: "GraphicsNDArray" = qattr((..., [1, 3]))
        br: "GraphicsNDArray" = qattr((..., [2, 0]))
        bg: "GraphicsNDArray" = qattr((..., [2, 1]))
        bb: "GraphicsNDArray" = qattr((..., [2, 2]))
        ba: "GraphicsNDArray" = qattr((..., [2, 3]))
        ar: "GraphicsNDArray" = qattr((..., [3, 0]))
        ag: "GraphicsNDArray" = qattr((..., [3, 1]))
        ab: "GraphicsNDArray" = qattr((..., [3, 2]))
        aa: "GraphicsNDArray" = qattr((..., [3, 3]))
        rrr: "GraphicsNDArray" = qattr((..., [0, 0, 0]))
        rrg: "GraphicsNDArray" = qattr((..., [0, 0, 1]))
        rrb: "GraphicsNDArray" = qattr((..., [0, 0, 2]))
        rra: "GraphicsNDArray" = qattr((..., [0, 0, 3]))
        rgr: "GraphicsNDArray" = qattr((..., [0, 1, 0]))
        rgg: "GraphicsNDArray" = qattr((..., [0, 1, 1]))
        rgb: "GraphicsNDArray" = qattr((..., [0, 1, 2]))
        rga: "GraphicsNDArray" = qattr((..., [0, 1, 3]))
        rbr: "GraphicsNDArray" = qattr((..., [0, 2, 0]))
        rbg: "GraphicsNDArray" = qattr((..., [0, 2, 1]))
        rbb: "GraphicsNDArray" = qattr((..., [0, 2, 2]))
        rba: "GraphicsNDArray" = qattr((..., [0, 2, 3]))
        rar: "GraphicsNDArray" = qattr((..., [0, 3, 0]))
        rag: "GraphicsNDArray" = qattr((..., [0, 3, 1]))
        rab: "GraphicsNDArray" = qattr((..., [0, 3, 2]))
        raa: "GraphicsNDArray" = qattr((..., [0, 3, 3]))
        grr: "GraphicsNDArray" = qattr((..., [1, 0, 0]))
        grg: "GraphicsNDArray" = qattr((..., [1, 0, 1]))
        grb: "GraphicsNDArray" = qattr((..., [1, 0, 2]))
        gra: "GraphicsNDArray" = qattr((..., [1, 0, 3]))
        ggr: "GraphicsNDArray" = qattr((..., [1, 1, 0]))
        ggg: "GraphicsNDArray" = qattr((..., [1, 1, 1]))
        ggb: "GraphicsNDArray" = qattr((..., [1, 1, 2]))
        gga: "GraphicsNDArray" = qattr((..., [1, 1, 3]))
        gbr: "GraphicsNDArray" = qattr((..., [1, 2, 0]))
        gbg: "GraphicsNDArray" = qattr((..., [1, 2, 1]))
        gbb: "GraphicsNDArray" = qattr((..., [1, 2, 2]))
        gba: "GraphicsNDArray" = qattr((..., [1, 2, 3]))
        gar: "GraphicsNDArray" = qattr((..., [1, 3, 0]))
        gag: "GraphicsNDArray" = qattr((..., [1, 3, 1]))
        gab: "GraphicsNDArray" = qattr((..., [1, 3, 2]))
        gaa: "GraphicsNDArray" = qattr((..., [1, 3, 3]))
        brr: "GraphicsNDArray" = qattr((..., [2, 0, 0]))
        brg: "GraphicsNDArray" = qattr((..., [2, 0, 1]))
        brb: "GraphicsNDArray" = qattr((..., [2, 0, 2]))
        bra: "GraphicsNDArray" = qattr((..., [2, 0, 3]))
        bgr: "GraphicsNDArray" = qattr((..., [2, 1, 0]))
        bgg: "GraphicsNDArray" = qattr((..., [2, 1, 1]))
        bgb: "GraphicsNDArray" = qattr((..., [2, 1, 2]))
        bga: "GraphicsNDArray" = qattr((..., [2, 1, 3]))
        bbr: "GraphicsNDArray" = qattr((..., [2, 2, 0]))
        bbg: "GraphicsNDArray" = qattr((..., [2, 2, 1]))
        bbb: "GraphicsNDArray" = qattr((..., [2, 2, 2]))
        bba: "GraphicsNDArray" = qattr((..., [2, 2, 3]))
        bar: "GraphicsNDArray" = qattr((..., [2, 3, 0]))
        bag: "GraphicsNDArray" = qattr((..., [2, 3, 1]))
        bab: "GraphicsNDArray" = qattr((..., [2, 3, 2]))
        baa: "GraphicsNDArray" = qattr((..., [2, 3, 3]))
        arr: "GraphicsNDArray" = qattr((..., [3, 0, 0]))
        arg: "GraphicsNDArray" = qattr((..., [3, 0, 1]))
        arb: "GraphicsNDArray" = qattr((..., [3, 0, 2]))
        ara: "GraphicsNDArray" = qattr((..., [3, 0, 3]))
        agr: "GraphicsNDArray" = qattr((..., [3, 1, 0]))
        agg: "GraphicsNDArray" = qattr((..., [3, 1, 1]))
        agb: "GraphicsNDArray" = qattr((..., [3, 1, 2]))
        aga: "GraphicsNDArray" = qattr((..., [3, 1, 3]))
        abr: "GraphicsNDArray" = qattr((..., [3, 2, 0]))
        abg: "GraphicsNDArray" = qattr((..., [3, 2, 1]))
        abb: "GraphicsNDArray" = qattr((..., [3, 2, 2]))
        aba: "GraphicsNDArray" = qattr((..., [3, 2, 3]))
        aar: "GraphicsNDArray" = qattr((..., [3, 3, 0]))
        aag: "GraphicsNDArray" = qattr((..., [3, 3, 1]))
        aab: "GraphicsNDArray" = qattr((..., [3, 3, 2]))
        aaa: "GraphicsNDArray" = qattr((..., [3, 3, 3]))
        rrrr: "GraphicsNDArray" = qattr((..., [0, 0, 0, 0]))
        rrrg: "GraphicsNDArray" = qattr((..., [0, 0, 0, 1]))
        rrrb: "GraphicsNDArray" = qattr((..., [0, 0, 0, 2]))
        rrra: "GraphicsNDArray" = qattr((..., [0, 0, 0, 3]))
        rrgr: "GraphicsNDArray" = qattr((..., [0, 0, 1, 0]))
        rrgg: "GraphicsNDArray" = qattr((..., [0, 0, 1, 1]))
        rrgb: "GraphicsNDArray" = qattr((..., [0, 0, 1, 2]))
        rrga: "GraphicsNDArray" = qattr((..., [0, 0, 1, 3]))
        rrbr: "GraphicsNDArray" = qattr((..., [0, 0, 2, 0]))
        rrbg: "GraphicsNDArray" = qattr((..., [0, 0, 2, 1]))
        rrbb: "GraphicsNDArray" = qattr((..., [0, 0, 2, 2]))
        rrba: "GraphicsNDArray" = qattr((..., [0, 0, 2, 3]))
        rrar: "GraphicsNDArray" = qattr((..., [0, 0, 3, 0]))
        rrag: "GraphicsNDArray" = qattr((..., [0, 0, 3, 1]))
        rrab: "GraphicsNDArray" = qattr((..., [0, 0, 3, 2]))
        rraa: "GraphicsNDArray" = qattr((..., [0, 0, 3, 3]))
        rgrr: "GraphicsNDArray" = qattr((..., [0, 1, 0, 0]))
        rgrg: "GraphicsNDArray" = qattr((..., [0, 1, 0, 1]))
        rgrb: "GraphicsNDArray" = qattr((..., [0, 1, 0, 2]))
        rgra: "GraphicsNDArray" = qattr((..., [0, 1, 0, 3]))
        rggr: "GraphicsNDArray" = qattr((..., [0, 1, 1, 0]))
        rggg: "GraphicsNDArray" = qattr((..., [0, 1, 1, 1]))
        rggb: "GraphicsNDArray" = qattr((..., [0, 1, 1, 2]))
        rgga: "GraphicsNDArray" = qattr((..., [0, 1, 1, 3]))
        rgbr: "GraphicsNDArray" = qattr((..., [0, 1, 2, 0]))
        rgbg: "GraphicsNDArray" = qattr((..., [0, 1, 2, 1]))
        rgbb: "GraphicsNDArray" = qattr((..., [0, 1, 2, 2]))
        rgba: "GraphicsNDArray" = qattr((..., [0, 1, 2, 3]))
        rgar: "GraphicsNDArray" = qattr((..., [0, 1, 3, 0]))
        rgag: "GraphicsNDArray" = qattr((..., [0, 1, 3, 1]))
        rgab: "GraphicsNDArray" = qattr((..., [0, 1, 3, 2]))
        rgaa: "GraphicsNDArray" = qattr((..., [0, 1, 3, 3]))
        rbrr: "GraphicsNDArray" = qattr((..., [0, 2, 0, 0]))
        rbrg: "GraphicsNDArray" = qattr((..., [0, 2, 0, 1]))
        rbrb: "GraphicsNDArray" = qattr((..., [0, 2, 0, 2]))
        rbra: "GraphicsNDArray" = qattr((..., [0, 2, 0, 3]))
        rbgr: "GraphicsNDArray" = qattr((..., [0, 2, 1, 0]))
        rbgg: "GraphicsNDArray" = qattr((..., [0, 2, 1, 1]))
        rbgb: "GraphicsNDArray" = qattr((..., [0, 2, 1, 2]))
        rbga: "GraphicsNDArray" = qattr((..., [0, 2, 1, 3]))
        rbbr: "GraphicsNDArray" = qattr((..., [0, 2, 2, 0]))
        rbbg: "GraphicsNDArray" = qattr((..., [0, 2, 2, 1]))
        rbbb: "GraphicsNDArray" = qattr((..., [0, 2, 2, 2]))
        rbba: "GraphicsNDArray" = qattr((..., [0, 2, 2, 3]))
        rbar: "GraphicsNDArray" = qattr((..., [0, 2, 3, 0]))
        rbag: "GraphicsNDArray" = qattr((..., [0, 2, 3, 1]))
        rbab: "GraphicsNDArray" = qattr((..., [0, 2, 3, 2]))
        rbaa: "GraphicsNDArray" = qattr((..., [0, 2, 3, 3]))
        rarr: "GraphicsNDArray" = qattr((..., [0, 3, 0, 0]))
        rarg: "GraphicsNDArray" = qattr((..., [0, 3, 0, 1]))
        rarb: "GraphicsNDArray" = qattr((..., [0, 3, 0, 2]))
        rara: "GraphicsNDArray" = qattr((..., [0, 3, 0, 3]))
        ragr: "GraphicsNDArray" = qattr((..., [0, 3, 1, 0]))
        ragg: "GraphicsNDArray" = qattr((..., [0, 3, 1, 1]))
        ragb: "GraphicsNDArray" = qattr((..., [0, 3, 1, 2]))
        raga: "GraphicsNDArray" = qattr((..., [0, 3, 1, 3]))
        rabr: "GraphicsNDArray" = qattr((..., [0, 3, 2, 0]))
        rabg: "GraphicsNDArray" = qattr((..., [0, 3, 2, 1]))
        rabb: "GraphicsNDArray" = qattr((..., [0, 3, 2, 2]))
        raba: "GraphicsNDArray" = qattr((..., [0, 3, 2, 3]))
        raar: "GraphicsNDArray" = qattr((..., [0, 3, 3, 0]))
        raag: "GraphicsNDArray" = qattr((..., [0, 3, 3, 1]))
        raab: "GraphicsNDArray" = qattr((..., [0, 3, 3, 2]))
        raaa: "GraphicsNDArray" = qattr((..., [0, 3, 3, 3]))
        grrr: "GraphicsNDArray" = qattr((..., [1, 0, 0, 0]))
        grrg: "GraphicsNDArray" = qattr((..., [1, 0, 0, 1]))
        grrb: "GraphicsNDArray" = qattr((..., [1, 0, 0, 2]))
        grra: "GraphicsNDArray" = qattr((..., [1, 0, 0, 3]))
        grgr: "GraphicsNDArray" = qattr((..., [1, 0, 1, 0]))
        grgg: "GraphicsNDArray" = qattr((..., [1, 0, 1, 1]))
        grgb: "GraphicsNDArray" = qattr((..., [1, 0, 1, 2]))
        grga: "GraphicsNDArray" = qattr((..., [1, 0, 1, 3]))
        grbr: "GraphicsNDArray" = qattr((..., [1, 0, 2, 0]))
        grbg: "GraphicsNDArray" = qattr((..., [1, 0, 2, 1]))
        grbb: "GraphicsNDArray" = qattr((..., [1, 0, 2, 2]))
        grba: "GraphicsNDArray" = qattr((..., [1, 0, 2, 3]))
        grar: "GraphicsNDArray" = qattr((..., [1, 0, 3, 0]))
        grag: "GraphicsNDArray" = qattr((..., [1, 0, 3, 1]))
        grab: "GraphicsNDArray" = qattr((..., [1, 0, 3, 2]))
        graa: "GraphicsNDArray" = qattr((..., [1, 0, 3, 3]))
        ggrr: "GraphicsNDArray" = qattr((..., [1, 1, 0, 0]))
        ggrg: "GraphicsNDArray" = qattr((..., [1, 1, 0, 1]))
        ggrb: "GraphicsNDArray" = qattr((..., [1, 1, 0, 2]))
        ggra: "GraphicsNDArray" = qattr((..., [1, 1, 0, 3]))
        gggr: "GraphicsNDArray" = qattr((..., [1, 1, 1, 0]))
        gggg: "GraphicsNDArray" = qattr((..., [1, 1, 1, 1]))
        gggb: "GraphicsNDArray" = qattr((..., [1, 1, 1, 2]))
        ggga: "GraphicsNDArray" = qattr((..., [1, 1, 1, 3]))
        ggbr: "GraphicsNDArray" = qattr((..., [1, 1, 2, 0]))
        ggbg: "GraphicsNDArray" = qattr((..., [1, 1, 2, 1]))
        ggbb: "GraphicsNDArray" = qattr((..., [1, 1, 2, 2]))
        ggba: "GraphicsNDArray" = qattr((..., [1, 1, 2, 3]))
        ggar: "GraphicsNDArray" = qattr((..., [1, 1, 3, 0]))
        ggag: "GraphicsNDArray" = qattr((..., [1, 1, 3, 1]))
        ggab: "GraphicsNDArray" = qattr((..., [1, 1, 3, 2]))
        ggaa: "GraphicsNDArray" = qattr((..., [1, 1, 3, 3]))
        gbrr: "GraphicsNDArray" = qattr((..., [1, 2, 0, 0]))
        gbrg: "GraphicsNDArray" = qattr((..., [1, 2, 0, 1]))
        gbrb: "GraphicsNDArray" = qattr((..., [1, 2, 0, 2]))
        gbra: "GraphicsNDArray" = qattr((..., [1, 2, 0, 3]))
        gbgr: "GraphicsNDArray" = qattr((..., [1, 2, 1, 0]))
        gbgg: "GraphicsNDArray" = qattr((..., [1, 2, 1, 1]))
        gbgb: "GraphicsNDArray" = qattr((..., [1, 2, 1, 2]))
        gbga: "GraphicsNDArray" = qattr((..., [1, 2, 1, 3]))
        gbbr: "GraphicsNDArray" = qattr((..., [1, 2, 2, 0]))
        gbbg: "GraphicsNDArray" = qattr((..., [1, 2, 2, 1]))
        gbbb: "GraphicsNDArray" = qattr((..., [1, 2, 2, 2]))
        gbba: "GraphicsNDArray" = qattr((..., [1, 2, 2, 3]))
        gbar: "GraphicsNDArray" = qattr((..., [1, 2, 3, 0]))
        gbag: "GraphicsNDArray" = qattr((..., [1, 2, 3, 1]))
        gbab: "GraphicsNDArray" = qattr((..., [1, 2, 3, 2]))
        gbaa: "GraphicsNDArray" = qattr((..., [1, 2, 3, 3]))
        garr: "GraphicsNDArray" = qattr((..., [1, 3, 0, 0]))
        garg: "GraphicsNDArray" = qattr((..., [1, 3, 0, 1]))
        garb: "GraphicsNDArray" = qattr((..., [1, 3, 0, 2]))
        gara: "GraphicsNDArray" = qattr((..., [1, 3, 0, 3]))
        gagr: "GraphicsNDArray" = qattr((..., [1, 3, 1, 0]))
        gagg: "GraphicsNDArray" = qattr((..., [1, 3, 1, 1]))
        gagb: "GraphicsNDArray" = qattr((..., [1, 3, 1, 2]))
        gaga: "GraphicsNDArray" = qattr((..., [1, 3, 1, 3]))
        gabr: "GraphicsNDArray" = qattr((..., [1, 3, 2, 0]))
        gabg: "GraphicsNDArray" = qattr((..., [1, 3, 2, 1]))
        gabb: "GraphicsNDArray" = qattr((..., [1, 3, 2, 2]))
        gaba: "GraphicsNDArray" = qattr((..., [1, 3, 2, 3]))
        gaar: "GraphicsNDArray" = qattr((..., [1, 3, 3, 0]))
        gaag: "GraphicsNDArray" = qattr((..., [1, 3, 3, 1]))
        gaab: "GraphicsNDArray" = qattr((..., [1, 3, 3, 2]))
        gaaa: "GraphicsNDArray" = qattr((..., [1, 3, 3, 3]))
        brrr: "GraphicsNDArray" = qattr((..., [2, 0, 0, 0]))
        brrg: "GraphicsNDArray" = qattr((..., [2, 0, 0, 1]))
        brrb: "GraphicsNDArray" = qattr((..., [2, 0, 0, 2]))
        brra: "GraphicsNDArray" = qattr((..., [2, 0, 0, 3]))
        brgr: "GraphicsNDArray" = qattr((..., [2, 0, 1, 0]))
        brgg: "GraphicsNDArray" = qattr((..., [2, 0, 1, 1]))
        brgb: "GraphicsNDArray" = qattr((..., [2, 0, 1, 2]))
        brga: "GraphicsNDArray" = qattr((..., [2, 0, 1, 3]))
        brbr: "GraphicsNDArray" = qattr((..., [2, 0, 2, 0]))
        brbg: "GraphicsNDArray" = qattr((..., [2, 0, 2, 1]))
        brbb: "GraphicsNDArray" = qattr((..., [2, 0, 2, 2]))
        brba: "GraphicsNDArray" = qattr((..., [2, 0, 2, 3]))
        brar: "GraphicsNDArray" = qattr((..., [2, 0, 3, 0]))
        brag: "GraphicsNDArray" = qattr((..., [2, 0, 3, 1]))
        brab: "GraphicsNDArray" = qattr((..., [2, 0, 3, 2]))
        braa: "GraphicsNDArray" = qattr((..., [2, 0, 3, 3]))
        bgrr: "GraphicsNDArray" = qattr((..., [2, 1, 0, 0]))
        bgrg: "GraphicsNDArray" = qattr((..., [2, 1, 0, 1]))
        bgrb: "GraphicsNDArray" = qattr((..., [2, 1, 0, 2]))
        bgra: "GraphicsNDArray" = qattr((..., [2, 1, 0, 3]))
        bggr: "GraphicsNDArray" = qattr((..., [2, 1, 1, 0]))
        bggg: "GraphicsNDArray" = qattr((..., [2, 1, 1, 1]))
        bggb: "GraphicsNDArray" = qattr((..., [2, 1, 1, 2]))
        bgga: "GraphicsNDArray" = qattr((..., [2, 1, 1, 3]))
        bgbr: "GraphicsNDArray" = qattr((..., [2, 1, 2, 0]))
        bgbg: "GraphicsNDArray" = qattr((..., [2, 1, 2, 1]))
        bgbb: "GraphicsNDArray" = qattr((..., [2, 1, 2, 2]))
        bgba: "GraphicsNDArray" = qattr((..., [2, 1, 2, 3]))
        bgar: "GraphicsNDArray" = qattr((..., [2, 1, 3, 0]))
        bgag: "GraphicsNDArray" = qattr((..., [2, 1, 3, 1]))
        bgab: "GraphicsNDArray" = qattr((..., [2, 1, 3, 2]))
        bgaa: "GraphicsNDArray" = qattr((..., [2, 1, 3, 3]))
        bbrr: "GraphicsNDArray" = qattr((..., [2, 2, 0, 0]))
        bbrg: "GraphicsNDArray" = qattr((..., [2, 2, 0, 1]))
        bbrb: "GraphicsNDArray" = qattr((..., [2, 2, 0, 2]))
        bbra: "GraphicsNDArray" = qattr((..., [2, 2, 0, 3]))
        bbgr: "GraphicsNDArray" = qattr((..., [2, 2, 1, 0]))
        bbgg: "GraphicsNDArray" = qattr((..., [2, 2, 1, 1]))
        bbgb: "GraphicsNDArray" = qattr((..., [2, 2, 1, 2]))
        bbga: "GraphicsNDArray" = qattr((..., [2, 2, 1, 3]))
        bbbr: "GraphicsNDArray" = qattr((..., [2, 2, 2, 0]))
        bbbg: "GraphicsNDArray" = qattr((..., [2, 2, 2, 1]))
        bbbb: "GraphicsNDArray" = qattr((..., [2, 2, 2, 2]))
        bbba: "GraphicsNDArray" = qattr((..., [2, 2, 2, 3]))
        bbar: "GraphicsNDArray" = qattr((..., [2, 2, 3, 0]))
        bbag: "GraphicsNDArray" = qattr((..., [2, 2, 3, 1]))
        bbab: "GraphicsNDArray" = qattr((..., [2, 2, 3, 2]))
        bbaa: "GraphicsNDArray" = qattr((..., [2, 2, 3, 3]))
        barr: "GraphicsNDArray" = qattr((..., [2, 3, 0, 0]))
        barg: "GraphicsNDArray" = qattr((..., [2, 3, 0, 1]))
        barb: "GraphicsNDArray" = qattr((..., [2, 3, 0, 2]))
        bara: "GraphicsNDArray" = qattr((..., [2, 3, 0, 3]))
        bagr: "GraphicsNDArray" = qattr((..., [2, 3, 1, 0]))
        bagg: "GraphicsNDArray" = qattr((..., [2, 3, 1, 1]))
        bagb: "GraphicsNDArray" = qattr((..., [2, 3, 1, 2]))
        baga: "GraphicsNDArray" = qattr((..., [2, 3, 1, 3]))
        babr: "GraphicsNDArray" = qattr((..., [2, 3, 2, 0]))
        babg: "GraphicsNDArray" = qattr((..., [2, 3, 2, 1]))
        babb: "GraphicsNDArray" = qattr((..., [2, 3, 2, 2]))
        baba: "GraphicsNDArray" = qattr((..., [2, 3, 2, 3]))
        baar: "GraphicsNDArray" = qattr((..., [2, 3, 3, 0]))
        baag: "GraphicsNDArray" = qattr((..., [2, 3, 3, 1]))
        baab: "GraphicsNDArray" = qattr((..., [2, 3, 3, 2]))
        baaa: "GraphicsNDArray" = qattr((..., [2, 3, 3, 3]))
        arrr: "GraphicsNDArray" = qattr((..., [3, 0, 0, 0]))
        arrg: "GraphicsNDArray" = qattr((..., [3, 0, 0, 1]))
        arrb: "GraphicsNDArray" = qattr((..., [3, 0, 0, 2]))
        arra: "GraphicsNDArray" = qattr((..., [3, 0, 0, 3]))
        argr: "GraphicsNDArray" = qattr((..., [3, 0, 1, 0]))
        argg: "GraphicsNDArray" = qattr((..., [3, 0, 1, 1]))
        argb: "GraphicsNDArray" = qattr((..., [3, 0, 1, 2]))
        arga: "GraphicsNDArray" = qattr((..., [3, 0, 1, 3]))
        arbr: "GraphicsNDArray" = qattr((..., [3, 0, 2, 0]))
        arbg: "GraphicsNDArray" = qattr((..., [3, 0, 2, 1]))
        arbb: "GraphicsNDArray" = qattr((..., [3, 0, 2, 2]))
        arba: "GraphicsNDArray" = qattr((..., [3, 0, 2, 3]))
        arar: "GraphicsNDArray" = qattr((..., [3, 0, 3, 0]))
        arag: "GraphicsNDArray" = qattr((..., [3, 0, 3, 1]))
        arab: "GraphicsNDArray" = qattr((..., [3, 0, 3, 2]))
        araa: "GraphicsNDArray" = qattr((..., [3, 0, 3, 3]))
        agrr: "GraphicsNDArray" = qattr((..., [3, 1, 0, 0]))
        agrg: "GraphicsNDArray" = qattr((..., [3, 1, 0, 1]))
        agrb: "GraphicsNDArray" = qattr((..., [3, 1, 0, 2]))
        agra: "GraphicsNDArray" = qattr((..., [3, 1, 0, 3]))
        aggr: "GraphicsNDArray" = qattr((..., [3, 1, 1, 0]))
        aggg: "GraphicsNDArray" = qattr((..., [3, 1, 1, 1]))
        aggb: "GraphicsNDArray" = qattr((..., [3, 1, 1, 2]))
        agga: "GraphicsNDArray" = qattr((..., [3, 1, 1, 3]))
        agbr: "GraphicsNDArray" = qattr((..., [3, 1, 2, 0]))
        agbg: "GraphicsNDArray" = qattr((..., [3, 1, 2, 1]))
        agbb: "GraphicsNDArray" = qattr((..., [3, 1, 2, 2]))
        agba: "GraphicsNDArray" = qattr((..., [3, 1, 2, 3]))
        agar: "GraphicsNDArray" = qattr((..., [3, 1, 3, 0]))
        agag: "GraphicsNDArray" = qattr((..., [3, 1, 3, 1]))
        agab: "GraphicsNDArray" = qattr((..., [3, 1, 3, 2]))
        agaa: "GraphicsNDArray" = qattr((..., [3, 1, 3, 3]))
        abrr: "GraphicsNDArray" = qattr((..., [3, 2, 0, 0]))
        abrg: "GraphicsNDArray" = qattr((..., [3, 2, 0, 1]))
        abrb: "GraphicsNDArray" = qattr((..., [3, 2, 0, 2]))
        abra: "GraphicsNDArray" = qattr((..., [3, 2, 0, 3]))
        abgr: "GraphicsNDArray" = qattr((..., [3, 2, 1, 0]))
        abgg: "GraphicsNDArray" = qattr((..., [3, 2, 1, 1]))
        abgb: "GraphicsNDArray" = qattr((..., [3, 2, 1, 2]))
        abga: "GraphicsNDArray" = qattr((..., [3, 2, 1, 3]))
        abbr: "GraphicsNDArray" = qattr((..., [3, 2, 2, 0]))
        abbg: "GraphicsNDArray" = qattr((..., [3, 2, 2, 1]))
        abbb: "GraphicsNDArray" = qattr((..., [3, 2, 2, 2]))
        abba: "GraphicsNDArray" = qattr((..., [3, 2, 2, 3]))
        abar: "GraphicsNDArray" = qattr((..., [3, 2, 3, 0]))
        abag: "GraphicsNDArray" = qattr((..., [3, 2, 3, 1]))
        abab: "GraphicsNDArray" = qattr((..., [3, 2, 3, 2]))
        abaa: "GraphicsNDArray" = qattr((..., [3, 2, 3, 3]))
        aarr: "GraphicsNDArray" = qattr((..., [3, 3, 0, 0]))
        aarg: "GraphicsNDArray" = qattr((..., [3, 3, 0, 1]))
        aarb: "GraphicsNDArray" = qattr((..., [3, 3, 0, 2]))
        aara: "GraphicsNDArray" = qattr((..., [3, 3, 0, 3]))
        aagr: "GraphicsNDArray" = qattr((..., [3, 3, 1, 0]))
        aagg: "GraphicsNDArray" = qattr((..., [3, 3, 1, 1]))
        aagb: "GraphicsNDArray" = qattr((..., [3, 3, 1, 2]))
        aaga: "GraphicsNDArray" = qattr((..., [3, 3, 1, 3]))
        aabr: "GraphicsNDArray" = qattr((..., [3, 3, 2, 0]))
        aabg: "GraphicsNDArray" = qattr((..., [3, 3, 2, 1]))
        aabb: "GraphicsNDArray" = qattr((..., [3, 3, 2, 2]))
        aaba: "GraphicsNDArray" = qattr((..., [3, 3, 2, 3]))
        aaar: "GraphicsNDArray" = qattr((..., [3, 3, 3, 0]))
        aaag: "GraphicsNDArray" = qattr((..., [3, 3, 3, 1]))
        aaab: "GraphicsNDArray" = qattr((..., [3, 3, 3, 2]))
        aaaa: "GraphicsNDArray" = qattr((..., [3, 3, 3, 3]))


NDArray = Union[numpy.ndarray, GraphicsNDArray]
T = TypeVar("T")

@type_match(numpy.ndarray)
def caster(arr):
    """
    Cast a ``numpy.ndarray`` into :py:class:`GraphicsNDArray`,
    but keeps other types unchanged.
    """
    return GraphicsNDArray(arr)


@type_match(GraphicsNDArray)
def backcaster(arr):
    """
    Cast a :py:class:`GraphicsNDArray` into ``numpy.ndarray``,
    but keeps other types unchanged.
    """
    return numpy.asarray(arr)


def treemap_backcaster(what):
    """
    Cast any :py:class:`GraphicsNDArray` in (nested) lists/dicts/tuples/sets
    into ``numpy.ndarray``, and keeps other types unchanged.
    """
    return container_catamorphism(what, backcaster)


def cast_graphics(func: T) -> T:
    """
    Decorator that casts all arguments into :py:class:`GraphicsNDArray`.
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        return treemap_backcaster(func(
            *(GraphicsNDArray(x) for x in args),
            **{k: GraphicsNDArray(v) for k, v in kwargs.items()}
        ))
    return wrapper


def treemap_cast_graphics(func: T) -> T:
    """
    Decorator that casts all arguments of type ``numpy.ndarray`` into :py:class:`GraphicsNDArray`.
    Also casts arrays in nested lists, dicts or tuples.
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        return treemap_backcaster(func(
            *container_catamorphism(args, caster),
            **container_catamorphism(kwargs, caster)
        ))
    return wrapper
